// Copyright (C) 2000 Open Source Telecom Corporation.
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#include "dbconfig.h"
#include "bayonnedb.h"
#include <getopt.h>
#include <pwd.h>

static int mainpid;
static bool running = false;
static Adapter **threads = NULL;
static unsigned adapter_count = 0;

static RETSIGTYPE final(int sig)
{
	int i;
	if(getpid() != mainpid)
	{
		kill(mainpid, sig);
		ccxx_sleep(~0);
	}

	signal(SIGINT, SIG_IGN);
	signal(SIGABRT, SIG_IGN);
	signal(SIGTERM, SIG_IGN);
	signal(SIGQUIT, SIG_IGN);

	if(running)
	{
		stopServers();
		finalize(sig);
	}

	delete[] threads;

	if(sig)
		slog(SLOG_WARNING) << "exiting: reason=" << sig << endl;
	else
		slog(SLOG_NOTICE) << "normal shutdown" << endl;
	exit(sig);
}

static void initial(int argc, char **argv)
{
	static struct option long_options[] = {
		{"background", 0, 0, 'D'},
		{"foreground", 0, 0, 'F'},
		{"version", 0, 0, 'V'},
		{"debug", 0, 0, 'x'},
		{"test", 0, 0, 'T'},
		{"initdb", 0, 0, 'I'},
		{0, 0, 0, 0}};

	static bool daemon = false;
	static bool usage = false;
	static bool test = false;

	int a = 0;
	char services[256];
	char drvpath[256];
	char *tok, *cp;
	int opt_index, opt;
	sigset_t sigs;
	struct passwd *pwd;

	mainpid = getpid();

	sigemptyset(&sigs);
	sigaddset(&sigs, SIGTERM);
	sigaddset(&sigs, SIGQUIT);
	sigaddset(&sigs, SIGINT);
	pthread_sigmask(SIG_BLOCK, &sigs, NULL);
	slog.level(SLOG_NOTICE);

	while(EOF != (opt = getopt_long(argc, argv, "TIFDVx", long_options, &opt_index)))
		switch(opt)
		{
		case 'I':
			initdb = true;
			break;
		case 'F':
			daemon = false;
			break;
		case 'D':
			daemon = true;
			break;
		case 'V':
			cout << VERPATH << endl;
			exit(0);
		case 'T':
			keyserver.setValue("sql", "../");
			test = true;
			daemon = false;
		case 'x':
			slog.level(SLOG_DEBUG);
			break;
		default:
			usage = true;
		}

	if(optind < argc || usage)
	{
		clog << "use: bayonnedb [-options]" << endl;
		exit(-1);
	}

	pwd = getpwnam(keyserver.getUsername());
	if(!pwd)
	{
		slog(SLOG_CRITICAL) << keyserver.getUsername() << ": invalid user account" << endl;
 		exit(-1);
	}
	setgid(pwd->pw_gid);
	setuid(pwd->pw_uid);

	endKeydata();
	endpwent();
	
	if(daemon)
	{
		close(0);
		close(1);
		close(2);
		pdetach();
		mainpid = getpid();
		slog.open("bayonnedb", SLOG_DAEMON);		
		slog(SLOG_NOTICE) << "daemon mode started" << endl;
	}

	signal(SIGABRT, final);
	try
	{
		adapter_count = keyserver.getThreads();
		threads = new Adapter*[adapter_count];
		while(a < adapter_count)
		{
			if(!a)
				threads[a++] = new Adapter(keyserver.getInterface(), keyserver.getPort());
			else
				threads[a++] = new Adapter(threads[0]);
		}		
		strcpy(services, keyserver.getServices());
		cp = strtok_r(services, " \t\n", &tok);
		if(cp)
		{
			slog(SLOG_DEBUG) << "loading " << keyserver.getBackend() << " driver" << endl;

			if(test)
				sprintf(drvpath, "../%s/db.%s",
					keyserver.getBackend(),
					keyserver.getBackend());
			else
				sprintf(drvpath, "%s/" VERPATH "/db.%s",
					keyserver.getDrivers(),
					keyserver.getBackend());
			new DSO(drvpath);
		}
		while(cp)
		{

			slog(SLOG_DEBUG) << "loading " << cp << " service" << endl;
			if(test)
				sprintf(drvpath, "./%s.service", cp);
			else
				sprintf(drvpath, "%s/" VERPATH "/%s.service",
					keyserver.getDrivers(), cp);
			new DSO(drvpath);

			slog(SLOG_DEBUG) << "loading " << cp << " backend" << endl;
			if(test)
				sprintf(drvpath, "../%s/%s.%s",
					keyserver.getBackend(),
					cp, keyserver.getBackend());
			else
				sprintf(drvpath, "%s/" VERPATH "/%s.%s",
					keyserver.getDrivers(),
					cp, keyserver.getBackend());

			new DSO(drvpath);
			cp = strtok_r(NULL, " \t\n", &tok);
		}
	}
	catch(DSO *dso)
	{
		slog(SLOG_CRITICAL) << dso->getError() << endl;
		final(-1);
	}
	catch(Socket *socket)
	{
		slog(SLOG_CRITICAL) << "unable to bind adapter" << endl;
		final(-1);
	}
	catch(Driver *driver)
	{
		slog(SLOG_CRITICAL) << "driver connection failed" << endl;
		final(-1);
	}
	catch(Database *database)
	{
		slog(SLOG_CRITICAL) << "database connection failed" << endl;
		final(-1);
	}
	initialize();
	startServers();
	running = true;

	signal(SIGTERM, final);
	signal(SIGQUIT, final);
	signal(SIGINT, final);
	signal(SIGABRT, final);
	pthread_sigmask(SIG_UNBLOCK, &sigs, NULL);
	slog("bayonnedb", SLOG_DAEMON);
	slog(SLOG_NOTICE) << "normal startup; " << keyserver.getConfig() << endl;
	scheduler();
}

KeyServer::KeyServer() :
Keydata("/bayonnedb/server")
{
	static KEYDEF defkeys[] = {
	{"config", "default runtime configuration"},
	{"drivers", "/usr/local/lib/bayonnedb"},
	{"sql", "/usr/local/share/bayonnedb"},
	{"backend", "postgres"},
	{"interface", "127.0.0.1"},
	{"port", "7002"},
	{"priority", "0"},
	{"threads", "5"},
	{"database", "bayonne"},
	{"user", "bayonne"},
	{"password", ""},
	{"interval", "60"},
	{"services", "cdr"},
	{NULL, NULL}};

	Load(defkeys);
}

char *KeyServer::getScript(const char *scriptfile)
{
	static char buffer[256];

	sprintf(buffer, "%s/%s/%s.sql", 
		getLast("sql"), getLast("backend"), scriptfile);
	return buffer;
}

InetAddress KeyServer::getInterface(void)
{
	return InetAddress(getLast("interface"));
}

int Adapter::count = 0;

Adapter::Adapter(InetAddress bind, tpport_t port) :
UDPSocket(bind, port), Server(keyserver.getPriority())
{
	id = ++count;
	sprintf(name, "adapter%d", id);
}

Adapter::Adapter(Adapter *bound) :
UDPSocket(*bound), Server(keyserver.getPriority())
{
	id = ++count;
	sprintf(name, "adapter%d", id);
}

Adapter::~Adapter()
{
	Terminate();
}

void Adapter::Run(void)
{
	unsigned char buffer[4096];
	struct sockaddr_in addr;
	socklen_t alen;
	int len;
	int reply = socket(AF_INET, SOCK_DGRAM, 17);

	shutdown(reply, 0);
	setCancel(THREAD_CANCEL_IMMEDIATE);

	for(;;)
	{
		alen = sizeof(addr);
		len = ::recvfrom(so, buffer, sizeof(buffer), 0, 
			(struct sockaddr *)&addr, &alen);
		if(len < 1)
			continue;
		preview(buffer, len);
		len = dispatch(buffer, len);
		if(len > 0)
			::sendto(reply, buffer, len, 0, 
				(struct sockaddr *)&addr, alen);
	}	
}

Server::Server(int pri) :
Thread(NULL, pri)
{
	next = first;
	first = this;
}

int main(int argc, char **argv)
{
	initial(argc, argv);
	scheduler();
	final(0);
}	
	
KeyServer keyserver;
bool initdb = false;




